@@ -8,6 +8,11 @@ import os
 import subprocess as sub
 import sys
 import tempfile
+import warnings
+import ConfigParser as cfgparser
+import logging
+import shutil
+import argparse
 
 try:
     from subprocess import check_output
@@ -24,37 +29,31 @@ except ImportError:
         Python 2.6.2
         '''
         process = sub.Popen(stdout=sub.PIPE, *popenargs, **kwargs)
-        output, unused_err = process.communicate()
+        output, _ = process.communicate()
         retcode = process.poll()
         if retcode:
             cmd = kwargs.get("args")
             if cmd is None:
                 cmd = popenargs[0]
-            error = sub.CalledProcessError(retcode, cmd)
-            error.output = output
-            raise error
+            err = sub.CalledProcessError(retcode, cmd)
+            err.output = output
+            raise err
         return output
 
     sub.check_output = backport_check_output
 
-__version__ = '0.2.4'
+__version__ = '0.3.0'
 
 BLOCK_SIZE = 4096
 
+NOT_IMPLEMENTED_MESSAGE = "This method isn't implemented for this backend!"
+
 
 def git(cliargs, *args, **kwargs):
     ''' Calls git commands with Popen arguments '''
     return sub.Popen(['git'] + cliargs, *args, **kwargs)
 
 
-def error(*args, **kwargs):
-    return print(*args, file=sys.stderr, **kwargs)
-
-
-def empty(*args, **kwargs):
-    pass
-
-
 def mkdir_p(path):
     import errno
     try:
@@ -95,10 +94,10 @@ def cat(instream, outstream):
     return cat_iter(readblocks(instream), outstream)
 
 
-def gitconfig_get(name, file=None):
+def gitconfig_get(name, cfgfile=None):
     args = ['config', '--get']
-    if file is not None:
-        args += ['--file', file]
+    if cfgfile is not None:
+        args += ['--file', cfgfile]
     args.append(name)
     p = git(args, stdout=sub.PIPE)
     output = p.communicate()[0].strip()
@@ -108,14 +107,33 @@ def gitconfig_get(name, file=None):
         return output
 
 
-def gitconfig_set(name, value, file=None):
+def gitconfig_set(name, value, cfgfile=None):
     args = ['git', 'config']
-    if file is not None:
-        args += ['--file', file]
+    if cfgfile is not None:
+        args += ['--file', cfgfile]
     args += [name, value]
     sub.check_call(args)
 
 
+def _config_path(path=None):
+    try:
+        root = sub.check_output('git rev-parse --show-toplevel'.split()).strip()
+    except sub.CalledProcessError:
+        raise RuntimeError('git-fat must be run from a git directory')
+    default_path = os.path.join(root, '.gitfat')
+    path = path or default_path
+    return path
+
+
+def _obj_dir():
+    try:
+        gitdir = sub.check_output('git rev-parse --git-dir'.split()).strip()
+    except sub.CalledProcessError:
+        raise RuntimeError('git-fat must be run from a git directory')
+    objdir = os.path.join(gitdir, 'fat', 'objects')
+    return objdir
+
+
 def http_get(baseurl, filename):
     ''' Returns file descriptor for http file stream, catches urllib2 errors '''
     import urllib2
@@ -125,43 +143,193 @@ def http_get(baseurl, filename):
         res = urllib2.urlopen(geturl)
         return res.fp
     except urllib2.URLError as e:
-        error("WARN: " + e.reason + ': {0}'.format(geturl))
+        logging.warning(e.reason + ': {0}'.format(geturl))
         return None
 
 
+def hash_stream(blockiter, outstream):
+    '''
+    Writes blockiter to outstream and returns the digest and bytes written
+    '''
+    hasher = hashlib.new('sha1')
+    bytes_written = 0
+
+    for block in blockiter:
+        # Add the block to be hashed
+        hasher.update(block)
+        bytes_written += len(block)
+        outstream.write(block)
+    outstream.flush()
+    return hasher.hexdigest(), bytes_written
+
+
+class BackendInterface(object):
+    """ __init__ and pull_files are required, push_files is optional """
+
+    def __init__(self, base_dir, **kwargs):
+        """ Configuration options should be set in here """
+        raise NotImplementedError(NOT_IMPLEMENTED_MESSAGE)
+
+    def push_files(self, file_list):
+        """ Return True if push was successful, False otherwise """
+        pass
+
+    def pull_files(self, file_list):
+        """ Return True if pull was successful, False otherwise """
+        raise NotImplementedError(NOT_IMPLEMENTED_MESSAGE)
+
+
+class CopyBackend(BackendInterface):
+    def __init__(self, base_dir, **kwargs):
+        other_path = kwargs.get('remote')
+        if not os.path.isdir(other_path):
+            raise RuntimeError('copybackend target path is not directory: {}'.format(other_path))
+
+        self.other_path = other_path
+        self.base_dir = base_dir
+
+    def pull_files(self, file_list):
+
+        for f in file_list:
+            fullpath = os.path.join(self.other_path, f)
+            shutil.copy2(fullpath, self.base_dir)
+
+    def push_files(self, file_list):
+
+        for f in file_list:
+            fullpath = os.path.join(self.base_dir, f)
+            shutil.copy2(fullpath, self.other_path)
+
+
+class HTTPBackend(BackendInterface):
+    """ Pull files from an HTTP server """
+
+    def __init__(self, base_dir, **kwargs):
+        remote_url = kwargs.get('remote')
+        if not remote_url:
+            raise RuntimeError('No remote url configured for http backend')
+
+        if not remote_url.startswith('http') or remote_url.startswith('https'):
+            raise RuntimeError('http remote url must start with http:// or https://')
+
+        self.remote_url = remote_url
+        self.base_dir = base_dir
+
+    def pull_files(self, file_list):
+        is_success = True
+
+        for o in file_list:
+            stream = http_get(self.remote_url, o)
+            blockiter = readblocks(stream)
+
+            # HTTP Error
+            if blockiter is None:
+                is_success = False
+                continue
+
+            fd, tmpname = tempfile.mkstemp(dir=self.base_dir)
+            with os.fdopen(fd, 'w') as tmpstream:
+                # Hash the input, write to temp file
+                digest, _ = hash_stream(blockiter, tmpstream)
+
+            if digest != o:
+                # Should I retry?
+                logging.error('Downloaded digest ({0}) did not match stored digest for orphan: {1}'.format(digest, o))
+                os.remove(tmpname)
+                is_success = False
+                continue
+
+            objfile = os.path.join(self.base_dir, digest)
+            os.chmod(tmpname, int('444', 8) & ~umask())
+            # Rename temp file
+            os.rename(tmpname, objfile)
+
+        return is_success
+
+
+class RSyncBackend(BackendInterface):
+    """ Push and pull files from rsync remote """
+
+    def __init__(self, base_dir, **kwargs):
+        remote_url = kwargs.get('remote')
+        ssh_user = kwargs.get('sshuser')
+        ssh_port = kwargs.get('sshport', '22')
+
+        if not remote_url:
+            raise RuntimeError("No remote url configured for rsync")
+
+        self.remote_url = remote_url
+        self.ssh_user = ssh_user
+        self.ssh_port = ssh_port
+        self.base_dir = base_dir
+
+    def _rsync(self, push):
+        ''' Construct the rsync command '''
+        cmd_tmpl = 'rsync --progress --ignore-existing --from0 --files-from=- {}/ {}/'
+        if push:
+            src, dst = self.base_dir, self.remote_url
+        else:
+            src, dst = self.remote_url, self.base_dir
+        cmd = cmd_tmpl.format(src, dst).split()
+
+        # extra must be passed in as single argv, which is why it's
+        # not in the template and split isn't called on it
+        extra = '--rsh=ssh'
+        if self.ssh_user:
+            extra = ' '.join([extra, '-l {}'.format(self.ssh_user)])
+        if self.ssh_port:
+            extra = ' '.join([extra, '-p {}'.format(self.ssh_port)])
+        cmd.append(extra)
+        return cmd
+
+    def pull_files(self, file_list):
+        rsync = self._rsync(push=False)
+        p = sub.Popen(rsync, stdin=sub.PIPE)
+        p.communicate(input='\x00'.join(file_list))
+        # TODO: fix for success check
+        return True
+
+    def push_files(self, file_list):
+        rsync = self._rsync(push=True)
+        p = sub.Popen(rsync, stdin=sub.PIPE)
+        p.communicate(input='\x00'.join(file_list))
+        # TODO: fix for success check
+        return True
+
+
+BACKEND_MAP = {
+    'rsync': RSyncBackend,
+    'http': HTTPBackend,
+    'copy': CopyBackend,
+}
+
+
 class GitFat(object):
 
-    def __init__(self):
-
-        def magiclen(enc_func):
-            return len(enc_func(hashlib.sha1('dummy').hexdigest(), 5))
+    def __init__(self, backend, full_history=False):
 
+        # The backend instance we use to get the files
+        self.backend = backend
+        self.full_history = full_history
+        self.rev = None  # Unused
+        self.objdir = _obj_dir()
         self._cookie = '#$# git-fat '
         self._format = self._cookie + '{digest} {size:20d}\n'
-        # Legacy format support below
+
+        # Legacy format support below, need to actually check the version once/if we have more than 2
         if os.environ.get('GIT_FAT_VERSION'):
             self._format = self._cookie + '{digest}\n'
 
-        self._magiclen = magiclen(self._encode)
+        # considers the git-fat version when generating the magic length
+        _ml = lambda fn: len(fn(hashlib.sha1('dummy').hexdigest(), 5))
+        self._magiclen = _ml(self._encode)
 
-    def configure(self, verbose=False, **kwargs):
+        self.configure()
+
+    def configure(self):
         '''
         Configure git-fat for usage: variables, environment
         '''
-
-        # Setup Variables
-        try:
-            self.gitroot = sub.check_output('git rev-parse --show-toplevel'.split()).strip()
-            self.gitdir = sub.check_output('git rev-parse --git-dir'.split()).strip()
-        except sub.CalledProcessError:
-            error('git-fat must be run from a git directory')
-            sys.exit(1)
-
-        self.objdir = os.path.join(self.gitdir, 'fat', 'objects')
-        self.cfgpath = os.path.join(self.gitroot, '.gitfat')
-
-        self.verbose = error if verbose or os.environ.get("GIT_FAT_VERBOSE") else empty
-
         if not self._configured():
             print('Setting filters in .git/config')
             gitconfig_set('filter.fat.clean', 'git-fat filter-clean %f')
@@ -178,66 +346,6 @@ class GitFat(object):
         filters = gitconfig_get('filter.fat.clean') and gitconfig_get('filter.fat.smudge')
         return filters and reqs
 
-    def _rsync_opts(self):
-        '''
-        Read rsync options from config
-        '''
-        if not os.path.isfile(self.cfgpath):
-            error('ERROR: git-fat requires that .gitfat is present to use rsync remotes')
-            sys.exit(1)
-
-        remote = gitconfig_get('rsync.remote', file=self.cfgpath)
-        if not remote:
-            error('ERROR: No rsync.remote in {0}'.format(self.cfgpath))
-            sys.exit(1)
-
-        ssh_port = gitconfig_get('rsync.sshport', file=self.cfgpath)
-        ssh_user = gitconfig_get('rsync.sshuser', file=self.cfgpath)
-        return remote, ssh_port, ssh_user
-
-    def _http_opts(self):
-        '''
-        Read http options from config
-        '''
-        if not os.path.isfile(self.cfgpath):
-            error('ERROR: git-fat requires that .gitfat is present to use http remotes')
-            sys.exit(1)
-
-        remote = gitconfig_get('http.remote', file=self.cfgpath)
-        if not remote:
-            error('ERROR: No http.remote in {0}'.format(self.cfgpath))
-            sys.exit(1)
-
-        if not remote.startswith('http') or remote.startswith('https'):
-            error('ERROR: http remote url must start with http:// or https://')
-            sys.exit(1)
-
-        return remote
-
-    def _rsync(self, push):
-        '''
-        Construct the rsync command
-        '''
-        (remote, ssh_port, ssh_user) = self._rsync_opts()
-        if push:
-            self.verbose('Pushing to %s' % (remote))
-        else:
-            self.verbose('Pulling from %s' % (remote))
-
-        cmd = ['rsync', '--progress', '--ignore-existing', '--from0', '--files-from=-']
-        rshopts = ''
-        if ssh_user:
-            rshopts += ' -l ' + ssh_user
-        if ssh_port:
-            rshopts += ' -p ' + ssh_port
-        if rshopts:
-            cmd.append('--rsh=ssh' + rshopts)
-        if push:
-            cmd += [self.objdir + '/', remote + '/']
-        else:
-            cmd += [remote + '/', self.objdir + '/']
-        return cmd
-
     def _encode(self, digest, size):
         '''
         Produce representation of file to be stored in repository. 20 characters can hold 64-bit integers.
@@ -263,25 +371,12 @@ class GitFat(object):
         # Put block back
         ret = prepend(block, stream_iter)
         if block.startswith(self._cookie):
-            assert(len(block) == self._magiclen)  # Sanity check
+            if len(block) != self._magiclen:  # Sanity check
+                warnings.warn('Found file with cookie but without magiclen')
+                return ret, False
             return ret, True
         return ret, False
 
-    def _hash_stream(self, blockiter, outstream):
-        '''
-        Writes blockiter to outstream and returns the digest and bytes written
-        '''
-        hasher = hashlib.new('sha1')
-        bytes_written = 0
-
-        for block in blockiter:
-            # Add the block to be hashed
-            hasher.update(block)
-            bytes_written += len(block)
-            outstream.write(block)
-        outstream.flush()
-        return hasher.hexdigest(), bytes_written
-
     def _get_digest(self, stream):
         '''
         Returns digest if stream is fatfile placeholder or '' if not
@@ -309,14 +404,14 @@ class GitFat(object):
         objs_dict = self._managed_files(**kwargs)
         return set(objs_dict.keys())
 
-    def _rev_list(self, rev=None, full_history=False):
+    def _rev_list(self):
         '''
         Generator for objects in rev. Returns (hash, type, size) tuple.
         '''
 
-        rev = rev or 'HEAD'
+        rev = self.rev or 'HEAD'
         # full_history implies --all
-        args = ['--all'] if full_history else ['--no-walk', rev]
+        args = ['--all'] if self.full_history else ['--no-walk', rev]
 
         # Get all the git objects in the current revision and in history if --all is specified
         revlist = git('rev-list --objects'.split() + args, stdout=sub.PIPE)
@@ -333,13 +428,13 @@ class GitFat(object):
 
         catfile.wait()
 
-    def _find_paths(self, hashes, rev=None, full_history=False):
+    def _find_paths(self, hashes):
         '''
         Takes a list of git object hashes and generates hash,path tuples
         '''
-        rev = rev or 'HEAD'
+        rev = self.rev or 'HEAD'
         # full_history implies --all
-        args = ['--all'] if full_history else ['--no-walk', rev]
+        args = ['--all'] if self.full_history else ['--no-walk', rev]
 
         revlist = git('rev-list --objects'.split() + args, stdout=sub.PIPE)
         for line in revlist.stdout:
@@ -355,11 +450,8 @@ class GitFat(object):
 
         revlist.wait()
 
-    def _managed_files(self, rev=None, full_history=False):
-        '''
-        Finds managed files in the specified revision
-        '''
-        revlistgen = self._rev_list(rev=rev, full_history=full_history)
+    def _managed_files(self):
+        revlistgen = self._rev_list()
         # Find any objects that are git-fat placeholders which are tracked in the repository
         managed = {}
         for objhash, objtype, size in revlistgen:
@@ -376,17 +468,18 @@ class GitFat(object):
         # filenames above, but was running into the memory buffer issue
         # Instead we just make another call to rev-list.  Takes more time, but still
         # only takes 5 seconds to traverse the entire history of a 22k commit repo
-        filedict = dict(self._find_paths(managed.keys(), rev=rev, full_history=full_history))
+        filedict = dict(self._find_paths(managed.keys()))
 
         # return a dict(git-fat hash -> filename)
         # git's objhash are the keys in `managed` and `filedict`
         ret = dict((j, filedict[i]) for i, j in managed.iteritems())
         return ret
 
-    def _orphan_files(self, patterns=[]):
+    def _orphan_files(self, patterns=None):
         '''
         generator for placeholders in working tree that match pattern
         '''
+        patterns = patterns or []
         # Null-terminated for proper file name handling (spaces)
         for fname in sub.check_output(['git', 'ls-files', '-z'] + patterns).split('\x00')[:-1]:
             stat = os.lstat(fname)
@@ -410,12 +503,12 @@ class GitFat(object):
             try:
                 with open(objfile) as f:
                     cat(f, outstream)
-                self.verbose('git-fat filter-smudge: restoring from %s' % objfile)
+                logging.info('git-fat filter-smudge: restoring from {}'.format(objfile))
             except IOError:
-                self.verbose('git-fat filter-smudge: fat object not found in cache %s' % objfile)
+                logging.info('git-fat filter-smudge: fat object not found in cache {}'.format(objfile))
                 outstream.write(block)
         else:
-            self.verbose('git-fat filter-smudge: not a managed file')
+            logging.info('git-fat filter-smudge: not a managed file')
             cat_iter(blockiter, sys.stdout)
 
     def _filter_clean(self, instream, outstream):
@@ -436,13 +529,13 @@ class GitFat(object):
         tmpstream = os.fdopen(fd, 'w')
 
         # Hash the input, write to temp file
-        digest, size = self._hash_stream(blockiter, tmpstream)
+        digest, size = hash_stream(blockiter, tmpstream)
         tmpstream.close()
 
         objfile = os.path.join(self.objdir, digest)
 
         if os.path.exists(objfile):
-            self.verbose('git-fat filter-clean: cached file already exists %s' % objfile)
+            logging.info('git-fat filter-clean: cached file already exists {}'.format(objfile))
             # Remove temp file
             os.remove(tmpname)
         else:
@@ -450,7 +543,7 @@ class GitFat(object):
             os.chmod(tmpname, int('444', 8) & ~umask())
             # Rename temp file
             os.rename(tmpname, objfile)
-            self.verbose('git-fat filter-clean: caching to %s' % objfile)
+            logging.info('git-fat filter-clean: caching to {}'.format(objfile))
 
         # Write placeholder to index
         outstream.write(self._encode(digest, size))
@@ -460,15 +553,15 @@ class GitFat(object):
         Public command to do the clean (should only be called by git)
         '''
         if cur_file and not self.can_clean_file(cur_file):
-            self.verbose(
+            logging.info(
                 "Not adding: {0}\n".format(cur_file) +
                 "It is not a new file and is not managed by git-fat"
             )
             # Git needs something, so we cat stdin to stdout
             cat(sys.stdin, sys.stdout)
         else:  # We clean the file
-            if (cur_file):
-                self.verbose("Adding {0}".format(cur_file))
+            if cur_file:
+                logging.info("Adding {0}".format(cur_file))
 
             self._filter_clean(sys.stdin, sys.stdout)
 
@@ -482,14 +575,14 @@ class GitFat(object):
         '''
         Find any files over size threshold in the repository.
         '''
-        revlistgen = self._rev_list(full_history=True)
+        revlistgen = self._rev_list()
         # Find any objects that are git-fat placeholders which are tracked in the repository
         objsizedict = {}
         for objhash, objtype, objsize in revlistgen:
             # files are of blob type
             if objtype == 'blob' and int(objsize) > size:
                 objsizedict[objhash] = objsize
-        for objhash, objpath in self._find_paths(objsizedict.keys(), full_history=True):
+        for objhash, objpath in self._find_paths(objsizedict.keys()):
             print(objhash, objsizedict[objhash], objpath)
 
     def _parse_ls_files(self, line):
@@ -499,65 +592,88 @@ class GitFat(object):
         filename = tail.strip()
         return mode, blobhash, stageno, filename
 
+    def _get_old_gitattributes(self):
+        """ Get the last .gitattributes file in HEAD, and return it """
+        ls_ga = git('ls-files -s .gitattributes'.split(), stdout=sub.PIPE)
+        lsout = ls_ga.stdout.read().strip()
+        ls_ga.wait()
+        if lsout:  # Always try to get the old gitattributes
+            ga_mode, ga_hash, ga_stno, _ = self._parse_ls_files(lsout)
+            ga_cat = git('cat-file blob {0}'.format(ga_hash).split(), stdout=sub.PIPE)
+            old_ga = ga_cat.stdout.read().splitlines()
+            ga_cat.wait()
+        else:
+            ga_mode, ga_stno, old_ga = '100644', '0', []
+        return old_ga, ga_mode, ga_stno
+
+    def _update_index(self, uip, mode, content, stageno, filename):
+        fmt = '{0} {1} {2}\t{3}\n'
+        uip.stdin.write(fmt.format(mode, content, stageno, filename))
+
+    def _add_gitattributes(self, newfiles, update_index):
+        """ Find the previous gitattributes file, and append to it """
+
+        old_ga, ga_mode, ga_stno = self._get_old_gitattributes()
+        ga_hashobj = git('hash-object -w --stdin'.split(), stdin=sub.PIPE,
+            stdout=sub.PIPE)
+        # Add lines to the .gitattributes file
+        new_ga = old_ga + ['{0} filter=fat -text'.format(f) for f in newfiles]
+        stdout, _ = ga_hashobj.communicate('\n'.join(new_ga) + '\n')
+        self._update_index(ga_mode, stdout.strip(), ga_stno, '.gitattributes')
+        return ga_mode, stdout.strip(), ga_stno, '.gitattributes'
+
+    def _process_index_filter_line(self, line, update_index, workdir, excludes):
+
+        mode, blobhash, stageno, filename = self._parse_ls_files(line)
+
+        if filename not in excludes or mode == "120000":
+            return None
+        # Save file to update .gitattributes
+        cleanedobj_hash = os.path.join(workdir, blobhash)
+        # if it hasn't already been cleaned
+        if not os.path.exists(cleanedobj_hash):
+            catfile = git('cat-file blob {0}'.format(blobhash).split(), stdout=sub.PIPE)
+            hashobj = git('hash-object -w --stdin'.split(), stdin=sub.PIPE, stdout=sub.PIPE)
+            self._filter_clean(catfile.stdout, hashobj.stdin)
+            hashobj.stdin.close()
+            objhash = hashobj.stdout.read().strip()
+            catfile.wait()
+            hashobj.wait()
+            with open(cleanedobj_hash, 'w') as cleaned:
+                cleaned.write(objhash + '\n')
+        else:
+            with open(cleanedobj_hash) as cleaned:
+                objhash = cleaned.read().strip()
+        return mode, objhash, stageno, filename
+
     def index_filter(self, filelist, add_gitattributes=True, **kwargs):
 
         workdir = os.path.join(self.gitdir, 'fat', 'index-filter')
         mkdir_p(workdir)
 
-        with open(filelist) as f:
-            files_to_exclude = f.read().splitlines()
+        with open(filelist) as excludes:
+            files_to_exclude = excludes.read().splitlines()
 
         ls_files = git('ls-files -s'.split(), stdout=sub.PIPE)
-        update_index = git('update-index --index-info'.split(), stdin=sub.PIPE)
-        lsfmt = '{0} {1} {2}\t{3}\n'
+        uip = git('update-index --index-info'.split(), stdin=sub.PIPE)
 
         newfiles = []
         for line in ls_files.stdout:
-            mode, blobhash, stageno, filename = self._parse_ls_files(line)
+            newfile = self._process_index_filter_line(line)
+            if newfile:
+                self._update_index(uip, *newfile)
+                # The filename is in the last position
+                newfiles.append(newfile[-1])
 
-            if filename not in files_to_exclude or mode == "120000":
-                continue
-            # Save file to update .gitattributes
-            newfiles.append(filename)
-            cleanedobj_hash = os.path.join(workdir, blobhash)
-            # if it hasn't already been cleaned
-            if not os.path.exists(cleanedobj_hash):
-                catfile = git('cat-file blob {0}'.format(blobhash).split(), stdout=sub.PIPE)
-                hashobj = git('hash-object -w --stdin'.split(), stdin=sub.PIPE, stdout=sub.PIPE)
-                self._filter_clean(catfile.stdout, hashobj.stdin)
-                hashobj.stdin.close()
-                objhash = hashobj.stdout.read().strip()
-                catfile.wait()
-                hashobj.wait()
-                with open(cleanedobj_hash, 'w') as f:
-                    f.write(objhash + '\n')
-            else:
-                with open(cleanedobj_hash) as f:
-                    objhash = f.read().strip()
-            # Write the placeholder to the index
-            update_index.stdin.write(lsfmt.format(mode, objhash, stageno, filename))
+        newfiles = self._filter_clean_index(ls_files.stdout, files_to_exclude)
 
         if add_gitattributes:
-            ls_ga = git('ls-files -s .gitattributes'.split(), stdout=sub.PIPE)
-            lsout = ls_ga.stdout.read().strip()
-            ls_ga.wait()
-            if lsout:  # Always try to get the old gitattributes
-                ga_mode, ga_hash, ga_stno, ga_filename = self._parse_ls_files(lsout)
-                ga_cat = git('cat-file blob {0}'.format(ga_hash).split(), stdout=sub.PIPE)
-                old_ga = ga_cat.stdout.read().splitlines()
-                ga_cat.wait()
-            else:
-                ga_mode, ga_stno, old_ga = '100644', '0', []
-            ga_hashobj = git('hash-object -w --stdin'.split(), stdin=sub.PIPE,
-                stdout=sub.PIPE)
-            new_ga = old_ga + ['{0} filter=fat -text'.format(f) for f in newfiles]
-            stdout, stderr = ga_hashobj.communicate('\n'.join(new_ga) + '\n')
-            update_index.stdin.write(lsfmt.format(ga_mode, stdout.strip(),
-                ga_stno, '.gitattributes'))
+            # Add the files to the gitattributes file and update the index
+            self._update_index(uip, *self._add_gitattributes(newfiles))
 
         ls_files.wait()
-        update_index.stdin.close()
-        update_index.wait()
+        uip.stdin.close()
+        uip.wait()
 
     def list_files(self, **kwargs):
         '''
@@ -608,7 +724,7 @@ class GitFat(object):
         # Flush the buffers to prevent deadlock from wait()
         # Caused when stdout from showfile is a large binary file and can't be fully buffered
         # I haven't figured out a way to avoid this unfortunately
-        for blk in blockiter:
+        for _ in blockiter:
             continue
 
         if showfile.wait() or is_fatfile:
@@ -620,37 +736,29 @@ class GitFat(object):
         return False
 
     def pull(self, pattern=None, **kwargs):
-        '''
-        Pull anything that I have referenced, but not stored
-        '''
+        """ Get orphans, call backend pull """
         cached_objs = self._cached_objects()
+
+        # TODO: Why use _orphan _and_ _referenced here?
         if pattern:
             # filter the working tree by a pattern
-            files = set(digest for digest, fname in self._orphan_files(patterns=[pattern])) - cached_objs
+            files = set(digest for digest, fname in self._orphan_files(patterns=(pattern,))) - cached_objs
         else:
             # default pull any object referenced but not stored
             files = self._referenced_objects(**kwargs) - cached_objs
 
-        if files:
-            print("Pulling: ", list(files))
-            rsync = self._rsync(push=False)
-            self.verbose('Executing: {0}'.format(rsync))
-            p = sub.Popen(rsync, stdin=sub.PIPE, preexec_fn=os.setsid)
-            p.communicate(input='\x00'.join(files))
-        else:
-            print("You've got everything! d(^_^)b")
-
+        if not self.backend.pull_files(files):
+            sys.exit(1)
+        # Make sure they're up to date
         self.checkout()
 
-    def push(self, **kwargs):
-        '''
-        Push anything that I have stored and referenced (rsync doesn't push if exists on remote)
-        '''
+    def push(self, pattern=None, **kwargs):
+        # We only want the intersection of the referenced files and ones we have cached
+        # Prevents file doesn't exist errors, while saving on bw by default (_referenced only
+        # checks HEAD for files)
         files = self._referenced_objects(**kwargs) & self._cached_objects()
-        rsync = self._rsync(push=True)
-        self.verbose('Executing: {0}'.format(rsync))
-        p = sub.Popen(rsync, stdin=sub.PIPE)
-        p.communicate(input='\x00'.join(files))
+        if not self.backend.push_files(files):
+            sys.exit(1)
 
     def _status(self, **kwargs):
         '''
@@ -676,141 +784,157 @@ class GitFat(object):
             for g in stale:
                 print('\t' + g)
 
-    def http_pull(self, **kwargs):
-        '''
-        Alternative to rsync (for anon clones)
-        '''
-        ret_code = 0
 
-        _, orphans = self._status(**kwargs)
-        baseurl = self._http_opts()
-        for o in orphans:
-            stream = http_get(baseurl, o)
-            blockiter = readblocks(stream)
+def _get_options(config, backend, cfg_file_path):
+    """ returns the options for a backend in dictionary form """
+    try:
+        opts = dict(config.items(backend))
+    except cfgparser.NoSectionError:
+        err = "No section found in {} for backend {}".format(cfg_file_path, backend)
+        raise RuntimeError(err)
+    return opts
 
-            # HTTP Error
-            if blockiter is None:
-                ret_code = 1
-                continue
 
-            fd, tmpname = tempfile.mkstemp(dir=self.objdir)
-            tmpstream = os.fdopen(fd, 'w')
+def _read_config(cfg_file_path=None):
+    config = cfgparser.SafeConfigParser()
+    if not os.path.exists(cfg_file_path):
+        raise RuntimeError("Configfile does not exist: {}".format(cfg_file_path))
+    try:
+        config.read(cfg_file_path)
+    except cfgparser.Error:  # TODO: figure out what to catch here
+        raise RuntimeError("Error reading or parsing configfile: {}".format(cfg_file_path))
+    return config
 
-            # Hash the input, write to temp file
-            digest, size = self._hash_stream(blockiter, tmpstream)
-            tmpstream.close()
 
-            if digest != o:
-                # Should I retry?
-                error('ERROR: Downloaded digest ({0}) did not match stored digest for orphan: {1}'.format(digest, o))
-                os.remove(tmpname)
-                ret_code = 1
-                continue
+def _parse_config(backend=None, cfg_file_path=None):
+    """ Parse the given config file and return the backend instance """
+    cfg_file_path = _config_path(path=cfg_file_path)
+    config = _read_config(cfg_file_path)
+    if backend is None:
+        try:
+            backends = config.sections()
+        except cfgparser.Error:
+            raise RuntimeError("Error reading or parsing configfile: {}".format(cfg_file_path))
+        if not backends:  # e.g. empty file
+            raise RuntimeError("No backends configured in config: {}".format(cfg_file_path))
+        backend = backends[0]
 
-            objfile = os.path.join(self.objdir, digest)
-            os.chmod(tmpname, int('444', 8) & ~umask())
-            # Rename temp file
-            os.rename(tmpname, objfile)
+    opts = _get_options(config, backend, cfg_file_path)
+    base_dir = _obj_dir()
 
-        self.checkout()
-        sys.exit(ret_code)
+    try:
+        Backend = BACKEND_MAP[backend]
+    except IndexError:
+        raise RuntimeError("Unknown backend specified: {}".format(backend))
+    return Backend(base_dir, **opts)
 
-    def http_push(self):
-        ''' NOT IMPLEMENTED '''
-        pass
+
+def _configure_logging(log_level):
+    log_format = '%(levelname)s:%(filename)s:%(message)s'
+    logging.basicConfig(format=log_format, level=log_level)
+
+
+def run(backend, **kwargs):
+    name = kwargs.pop('func')
+    full_history = kwargs.pop('full_history')
+    gitfat = GitFat(backend, full_history=full_history)
+    fn = name.replace("-", "_")
+    assert hasattr(gitfat, fn), "Unknown function called!"
+    getattr(gitfat, fn)(**kwargs)
 
 
 def main():
 
-    import argparse
-
-    if sys.argv[1] in [ c + 'version' for c in '','-','--']:
-        print(__version__)
-        sys.exit(0)
-
-    fat = GitFat()
-
-    parser = argparse.ArgumentParser(usage=("%(prog)s [global options] command [command options]\n"
-        "       %(prog)s -h for full usage."))  # Six spaces for len('usage: ')
+    parser = argparse.ArgumentParser(argument_default=argparse.SUPPRESS,
+        description='A tool for managing large binary files in git repositories.')
     subparser = parser.add_subparsers()
 
-    parser.add_argument('-a', "--full-history", dest='full_history', action='store_true',
+    # Global options
+    parser.add_argument('-a', "--full-history", dest='full_history', action='store_true', default=False,
         help='Look for git-fat placeholder files in the entire history instead of just the working copy')
     parser.add_argument('-v', "--verbose", dest='verbose', action='store_true',
-        help='Verbose output')
+        help='Get verbose output about what git-fat is doing')
+    parser.add_argument('-d', "--debug", dest='debug', action='store_true',
+        help='Get debugging output about what git-fat is doing')
+    parser.add_argument('-c', "--config", dest='config_file', type=str,
+        help='Specify which config file to use (defaults to .gitfat)')
 
-    # Empty function for legacy api; config gets called every time
+    # redundant function for legacy api; config gets called every time.
     # (assuming if user is calling git-fat they want it configured)
-    parser_init = subparser.add_parser('init',
-        help='Initialize git-fat')
-    parser_init.set_defaults(func=empty)
+    # plus people like running init when setting things up d(^_^)b
+    sp = subparser.add_parser('init', help='Initialize git-fat')
+    sp.set_defaults(func="configure")
 
-    parser_filter_clean = subparser.add_parser('filter-clean',
-        help='filter-clean to be called only by git')
-    parser_filter_clean.add_argument("cur_file", nargs="?")
-    parser_filter_clean.set_defaults(func=fat.filter_clean)
+    sp = subparser.add_parser('filter-clean', help="Internal function used by git")
+    sp.add_argument("cur_file", nargs="?")
+    sp.set_defaults(func='filter_clean')
 
-    parser_filter_smudge = subparser.add_parser('filter-smudge',
-        help='filter-smudge to be called only by git')
-    parser_filter_smudge.add_argument("cur_file", nargs="?")  # Currently unused
-    parser_filter_smudge.set_defaults(func=fat.filter_smudge)
+    sp = subparser.add_parser('filter-smudge', help="Internal function used by git")
+    sp.add_argument("cur_file", nargs="?")
+    sp.set_defaults(func='filter_smudge')
 
-    parser_push = subparser.add_parser('push',
-        help='push cache to remote git-fat server')
-    parser_push.set_defaults(func=fat.push)
+    sp = subparser.add_parser('push', help='push cache to remote git-fat server')
+    sp.add_argument("backend", nargs="?", help='pull using given backend')
+    sp.set_defaults(func='push')
 
-    parser_pull = subparser.add_parser('pull',
-        help='pull fatfiles from remote git-fat server')
-    parser_pull.add_argument("pattern", nargs="?",
-        help='pull only files matching pattern')
-    parser_pull.set_defaults(func=fat.pull)
+    sp = subparser.add_parser('pull', help='pull fatfiles from remote git-fat server')
+    sp.add_argument("backend", nargs="?", help='pull using given backend')
+    sp.set_defaults(func='pull')
 
-    parser_checkout = subparser.add_parser('checkout',
-        help='resmudge all orphan objects')
-    parser_checkout.set_defaults(func=fat.checkout)
+    sp = subparser.add_parser('checkout', help='resmudge all orphan objects')
+    sp.set_defaults(func='checkout')
 
-    parser_find = subparser.add_parser('find',
-        help='find all objects over [size]')
-    parser_find.add_argument("size", type=int,
-        help='threshold size in bytes')
-    parser_find.set_defaults(func=fat.find)
+    sp = subparser.add_parser('find', help='find all objects over [size]')
+    sp.add_argument("size", type=int, help='threshold size in bytes')
+    sp.set_defaults(func='find')
 
-    parser_status = subparser.add_parser('status',
-        help='print orphan and stale objects')
-    parser_status.set_defaults(func=fat.status)
+    sp = subparser.add_parser('status', help='print orphan and stale objects')
+    sp.set_defaults(func='status')
 
-    parser_list = subparser.add_parser('list',
-        help='list all files managed by git-fat')
-    parser_list.set_defaults(func=fat.list_files)
+    sp = subparser.add_parser('list', help='list all files managed by git-fat')
+    sp.set_defaults(func='list_files')
 
-    parser_list = subparser.add_parser('pull-http',
-        help='anonymously download git-fat files over http')
-    parser_list.set_defaults(func=fat.http_pull)
+    # Legacy function to preserve backwards compatability
+    sp = subparser.add_parser('pull-http', help="Deprecated, use `pull http` (no dash) instead")
+    sp.set_defaults(func='pull', backend='http')
 
-    parser_index_filter = subparser.add_parser('index-filter',
-        help='git fat index-filter for filter-branch')
-    parser_index_filter.add_argument('filelist',
-        help='file containing all files to import to git-fat')
-    parser_index_filter.add_argument('-x', dest='add_gitattributes',
+    sp = subparser.add_parser('index-filter', help='git fat index-filter for filter-branch')
+    sp.add_argument('filelist', help='file containing all files to import to git-fat')
+    sp.add_argument('-x', dest='add_gitattributes',
         help='prevent adding excluded to .gitattributes', action='store_false')
-    parser_index_filter.set_defaults(func=fat.index_filter)
+    sp.set_defaults(func='index_filter')
+
+    try:
+        # Being lazy by not using argparse
+        if sys.argv[1] in [c + 'version' for c in '', '-', '--']:
+            print(__version__)
+            sys.exit(0)
+    except IndexError:
+        parser.print_help()
+        sys.exit(1)
 
     args = parser.parse_args()
-
     kwargs = dict(vars(args))
-    del kwargs['func']
 
-    # First configure git-fat
-    fat.configure(**kwargs)
+    if kwargs.pop('debug', None):
+        log_level = logging.DEBUG
+    elif kwargs.pop('verbose', None):
+        log_level = logging.INFO
+    else:
+        log_level = logging.WARNING
+    _configure_logging(log_level)
 
-    # don't need this after configuring, breaks functions because of kwargs
-    del kwargs['verbose']
-    # Then execute function
     try:
-        args.func(**kwargs)
+        backend_opt = kwargs.pop('backend', None)
+        config_file = kwargs.pop('config_file', None)
+        backend = _parse_config(backend=backend_opt, cfg_file_path=config_file)
+        run(backend, **kwargs)
+    except RuntimeError as err:
+        logging.error(str(err))
+        sys.exit(1)
     except:
         if kwargs.get('cur_file'):
-            fat.verbose("ERROR: processing file: " + kwargs.get('cur_file'))
+            logging.error("processing file: " + kwargs.get('cur_file'))
         raise
 
 
